

CIV_DEPRECATION_MESSAGE = "The word 'CIVILIAN' has been replaced with 'SUBJECT' in " +\
    f"TableType names to be consistent with the Stanford Open Policing Project "+\
    f"(https://github.com/stanford-policylab/opp/blob/master/data_readme.md). "

CIV_REPLACEMENT_MESSAGE =  "{} is being corrected to {}. "+\
                    "Please replace 'CIVILIAN' with 'SUBJECT' in your code as this correction will be removed "+\
                    "in a future release."